# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import io
import logging
import sys

from collections.abc import Sequence
from typing import Optional, List, TextIO

from pyomo.common.config import (
    ConfigDict,
    ConfigValue,
    NonNegativeFloat,
    NonNegativeInt,
    Bool,
    Path,
    document_configdict,
)
from pyomo.common.log import LogStream
from pyomo.common.numeric_types import native_logical_types
from pyomo.common.timing import HierarchicalTimer


def TextIO_or_Logger(val):
    """Validates and converts input into a list of valid output streams.

    Accepts:
      - :obj:`sys.stdout`
      - :class:`io.TextIOBase`
      - :class:`logging.Logger`
      - ``True`` (alias for :obj:`sys.stdout`)

    Returns
    -------
    List[io.TextIOBase]
        A list of validated output streams.

    Raises
    ------
    ValueError
        If an invalid type is provided.

    """
    if isinstance(val, Sequence) and not isinstance(val, (str, bytes)):
        val = list(val)
    else:
        val = [val]

    ans = []
    for v in val:
        if v.__class__ in native_logical_types:
            if v:
                ans.append(sys.stdout)
        elif isinstance(v, (sys.stdout.__class__, io.TextIOBase)):
            # We are guarding against file-like classes that do not derive from
            # TextIOBase but are assigned to stdout / stderr.
            # We still want to accept those classes.
            ans.append(v)
        elif isinstance(v, logging.Logger):
            ans.append(LogStream(level=logging.INFO, logger=v))
        else:
            raise ValueError(
                f"Expected sys.stdout, io.TextIOBase, Logger, or bool, but received {v.__class__}"
            )
    return ans


@document_configdict()
class SolverConfig(ConfigDict):
    """
    Common configuration options for all solver interfaces
    """

    def __init__(
        self,
        description=None,
        doc=None,
        implicit=False,
        implicit_domain=None,
        visibility=0,
    ):
        super().__init__(
            description=description,
            doc=doc,
            implicit=implicit,
            implicit_domain=implicit_domain,
            visibility=visibility,
        )

        self.tee: List[TextIO] = self.declare(
            'tee',
            ConfigValue(
                domain=TextIO_or_Logger,
                default=False,
                description="""``tee`` accepts :py:class:`bool`,
                :py:class:`io.TextIOBase`, or :py:class:`logging.Logger`
                (or a list of these types).  ``True`` is mapped to
                ``sys.stdout``.  The solver log will be printed to each of
                these streams / destinations.""",
            ),
        )
        self.working_dir: Optional[Path] = self.declare(
            'working_dir',
            ConfigValue(
                domain=Path(),
                default=None,
                description="The directory in which generated files should be saved. "
                "This replaces the `keepfiles` option.",
            ),
        )
        self.load_solutions: bool = self.declare(
            'load_solutions',
            ConfigValue(
                domain=Bool,
                default=True,
                description="If True, the values of the primal variables will be loaded into the model.",
            ),
        )
        self.raise_exception_on_nonoptimal_result: bool = self.declare(
            'raise_exception_on_nonoptimal_result',
            ConfigValue(
                domain=Bool,
                default=True,
                description="If False, the `solve` method will continue processing "
                "even if the returned result is nonoptimal.",
            ),
        )
        self.symbolic_solver_labels: bool = self.declare(
            'symbolic_solver_labels',
            ConfigValue(
                domain=Bool,
                default=False,
                description="If True, the names given to the solver will reflect the names of the Pyomo components. "
                "Cannot be changed after set_instance is called.",
            ),
        )
        self.timer: Optional[HierarchicalTimer] = self.declare(
            'timer',
            ConfigValue(
                default=None,
                description="A timer object for recording relevant process timing data.",
            ),
        )
        self.threads: Optional[int] = self.declare(
            'threads',
            ConfigValue(
                domain=NonNegativeInt,
                description="Number of threads to be used by a solver.",
                default=None,
            ),
        )
        self.time_limit: Optional[float] = self.declare(
            'time_limit',
            ConfigValue(
                domain=NonNegativeFloat,
                description="Time limit applied to the solver (in seconds).",
            ),
        )
        self.solver_options: ConfigDict = self.declare(
            'solver_options',
            ConfigDict(implicit=True, description="Options to pass to the solver."),
        )


@document_configdict()
class BranchAndBoundConfig(SolverConfig):
    """Base config for all direct MIP solver interfaces"""

    def __init__(
        self,
        description=None,
        doc=None,
        implicit=False,
        implicit_domain=None,
        visibility=0,
    ):
        super().__init__(
            description=description,
            doc=doc,
            implicit=implicit,
            implicit_domain=implicit_domain,
            visibility=visibility,
        )

        self.rel_gap: Optional[float] = self.declare(
            'rel_gap',
            ConfigValue(
                domain=NonNegativeFloat,
                description="Optional termination condition; the relative value of the "
                "gap in relation to the best bound",
            ),
        )
        self.abs_gap: Optional[float] = self.declare(
            'abs_gap',
            ConfigValue(
                domain=NonNegativeFloat,
                description="Optional termination condition; the absolute value of the "
                "difference between the incumbent and best bound",
            ),
        )


@document_configdict()
class AutoUpdateConfig(ConfigDict):
    """Control which parts of the model are automatically checked and/or
    updated upon re-solve

    """

    def __init__(
        self,
        description=None,
        doc=None,
        implicit=False,
        implicit_domain=None,
        visibility=0,
    ):
        if doc is None:
            doc = 'Configuration options to detect changes in model between solves'
        super().__init__(
            description=description,
            doc=doc,
            implicit=implicit,
            implicit_domain=implicit_domain,
            visibility=visibility,
        )

        self.check_for_new_or_removed_constraints: bool = self.declare(
            'check_for_new_or_removed_constraints',
            ConfigValue(
                domain=bool,
                default=True,
                description="""
                If False, new/old constraints will not be automatically detected on subsequent
                solves. Use False only when manually updating the solver with opt.add_constraints()
                and opt.remove_constraints() or when you are certain constraints are not being
                added to/removed from the model.""",
            ),
        )
        self.check_for_new_or_removed_vars: bool = self.declare(
            'check_for_new_or_removed_vars',
            ConfigValue(
                domain=bool,
                default=True,
                description="""
                If False, new/old variables will not be automatically detected on subsequent 
                solves. Use False only when manually updating the solver with opt.add_variables() and 
                opt.remove_variables() or when you are certain variables are not being added to /
                removed from the model.""",
            ),
        )
        self.check_for_new_or_removed_params: bool = self.declare(
            'check_for_new_or_removed_params',
            ConfigValue(
                domain=bool,
                default=True,
                description="""
                If False, new/old parameters will not be automatically detected on subsequent 
                solves. Use False only when manually updating the solver with opt.add_parameters() and 
                opt.remove_parameters() or when you are certain parameters are not being added to /
                removed from the model.""",
            ),
        )
        self.check_for_new_objective: bool = self.declare(
            'check_for_new_objective',
            ConfigValue(
                domain=bool,
                default=True,
                description="""
                If False, new/old objectives will not be automatically detected on subsequent 
                solves. Use False only when manually updating the solver with opt.set_objective() or 
                when you are certain objectives are not being added to / removed from the model.""",
            ),
        )
        self.update_constraints: bool = self.declare(
            'update_constraints',
            ConfigValue(
                domain=bool,
                default=True,
                description="""
                If False, changes to existing constraints will not be automatically detected on 
                subsequent solves. This includes changes to the lower, body, and upper attributes of 
                constraints. Use False only when manually updating the solver with 
                opt.remove_constraints() and opt.add_constraints() or when you are certain constraints 
                are not being modified.""",
            ),
        )
        self.update_vars: bool = self.declare(
            'update_vars',
            ConfigValue(
                domain=bool,
                default=True,
                description="""
                If False, changes to existing variables will not be automatically detected on 
                subsequent solves. This includes changes to the lb, ub, domain, and fixed 
                attributes of variables. Use False only when manually updating the solver with 
                opt.update_variables() or when you are certain variables are not being modified.""",
            ),
        )
        self.update_parameters: bool = self.declare(
            'update_parameters',
            ConfigValue(
                domain=bool,
                default=True,
                description="""
                If False, changes to parameter values will not be automatically detected on 
                subsequent solves. Use False only when manually updating the solver with 
                opt.update_parameters() or when you are certain parameters are not being modified.""",
            ),
        )
        self.update_named_expressions: bool = self.declare(
            'update_named_expressions',
            ConfigValue(
                domain=bool,
                default=True,
                description="""
                If False, changes to Expressions will not be automatically detected on 
                subsequent solves. Use False only when manually updating the solver with 
                opt.remove_constraints() and opt.add_constraints() or when you are certain 
                Expressions are not being modified.""",
            ),
        )
        self.update_objective: bool = self.declare(
            'update_objective',
            ConfigValue(
                domain=bool,
                default=True,
                description="""
                If False, changes to objectives will not be automatically detected on 
                subsequent solves. This includes the expr and sense attributes of objectives. Use 
                False only when manually updating the solver with opt.set_objective() or when you are 
                certain objectives are not being modified.""",
            ),
        )


@document_configdict()
class PersistentSolverConfig(SolverConfig):
    """Base config for all persistent solver interfaces"""

    def __init__(
        self,
        description=None,
        doc=None,
        implicit=False,
        implicit_domain=None,
        visibility=0,
    ):
        super().__init__(
            description=description,
            doc=doc,
            implicit=implicit,
            implicit_domain=implicit_domain,
            visibility=visibility,
        )

        self.auto_updates: AutoUpdateConfig = self.declare(
            'auto_updates', AutoUpdateConfig()
        )


@document_configdict()
class PersistentBranchAndBoundConfig(PersistentSolverConfig, BranchAndBoundConfig):
    """Base config for all persistent MIP solver interfaces"""

    def __init__(
        self,
        description=None,
        doc=None,
        implicit=False,
        implicit_domain=None,
        visibility=0,
    ):
        super().__init__(
            description=description,
            doc=doc,
            implicit=implicit,
            implicit_domain=implicit_domain,
            visibility=visibility,
        )



# -*- coding: utf-8 -*-
# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
"""Tests infeasible model debugging utilities."""

import logging

from io import StringIO

import pyomo.common.unittest as unittest
from pyomo.common.log import LoggingIntercept
from pyomo.environ import ConcreteModel, Constraint, Var, inequality
from pyomo.util.infeasible import (
    log_active_constraints,
    log_close_to_bounds,
    log_infeasible_bounds,
    log_infeasible_constraints,
)


class TestInfeasible(unittest.TestCase):
    """Tests infeasible model debugging utilities."""

    def build_model(self):
        m = ConcreteModel()
        m.x = Var(initialize=1)
        m.c1 = Constraint(expr=m.x >= 2)
        m.c2 = Constraint(expr=m.x == 4)
        m.c3 = Constraint(expr=m.x <= 0)
        m.y = Var(bounds=(0, 2), initialize=1.9999999)
        m.c4 = Constraint(expr=m.y >= m.y.value)
        m.z = Var(bounds=(0, 6))
        m.c5 = Constraint(expr=inequality(5, m.z, 10), doc="Range infeasible")
        m.c6 = Constraint(expr=m.x + m.y <= 6, doc="Feasible")
        m.c7 = Constraint(expr=m.z == 6, doc="Equality infeasible")
        m.c8 = Constraint(expr=inequality(3, m.x, 6), doc="Range lb infeasible")
        m.c9 = Constraint(expr=inequality(0, m.x, 0.5), doc="Range ub infeasible")
        m.c10 = Constraint(expr=m.y >= 3, doc="Inactive")
        m.c10.deactivate()
        m.c11 = Constraint(expr=m.y <= m.y.value)
        m.yy = Var(bounds=(0, 1), initialize=1e-7, doc="Close to lower bound")
        m.y3 = Var(bounds=(0, 1e-7), initialize=0, doc="Bounds too close")
        m.y4 = Var(bounds=(0, 1), initialize=2, doc="Fixed out of bounds.")
        m.y4.fix()
        return m

    def test_log_infeasible_constraints(self):
        """Test for logging of infeasible constraints."""
        m = self.build_model()

        with LoggingIntercept(None, 'pyomo.util.infeasible') as LOG:
            log_infeasible_constraints(m)
        self.assertEqual(
            'log_infeasible_constraints() called with a logger whose '
            'effective level is higher than logging.INFO: no output '
            'will be logged regardless of constraint feasibility',
            LOG.getvalue().strip(),
        )

        output = StringIO()
        with LoggingIntercept(output, 'pyomo.util.infeasible', logging.INFO):
            log_infeasible_constraints(m)
        expected_output = [
            "CONSTR c1: 2.0 </= 1",
            "CONSTR c2: 1 =/= 4.0",
            "CONSTR c3: 1 </= 0.0",
            "CONSTR c5: 5.0 <?= evaluation error <?= 10.0",
            "CONSTR c7: evaluation error =?= 6.0",
            "CONSTR c8: 3.0 </= 1 <= 6.0",
            "CONSTR c9: 0.0 <= 1 </= 0.5",
        ]
        self.assertEqual(expected_output, output.getvalue().splitlines())

    def test_log_infeasible_bounds(self):
        """Test for logging of infeasible variable bounds."""
        m = self.build_model()
        m.x.setlb(2)
        m.x.setub(0)

        with LoggingIntercept(None, 'pyomo.util.infeasible') as LOG:
            log_infeasible_bounds(m)
        self.assertEqual(
            'log_infeasible_bounds() called with a logger whose '
            'effective level is higher than logging.INFO: no output '
            'will be logged regardless of bound feasibility',
            LOG.getvalue().strip(),
        )

        output = StringIO()
        with LoggingIntercept(output, 'pyomo.util', logging.INFO):
            log_infeasible_bounds(m)
        expected_output = [
            "VAR x: LB 2 </= 1",
            "VAR x: 1 </= UB 0",
            "VAR z: no assigned value.",
            "VAR y4: 2 </= UB 1",
        ]
        self.assertEqual(expected_output, output.getvalue().splitlines())

    def test_log_active_constraints(self):
        """Test for logging of active constraints."""
        m = self.build_model()
        depr = StringIO()
        output = StringIO()
        with LoggingIntercept(depr, 'pyomo.util', logging.WARNING):
            log_active_constraints(m)
        self.assertIn("log_active_constraints is deprecated.", depr.getvalue())
        with LoggingIntercept(output, 'pyomo.util', logging.INFO):
            log_active_constraints(m)
        expected_output = [
            "c1 active",
            "c2 active",
            "c3 active",
            "c4 active",
            "c5 active",
            "c6 active",
            "c7 active",
            "c8 active",
            "c9 active",
            "c11 active",
        ]
        self.assertEqual(
            expected_output, output.getvalue()[len(depr.getvalue()) :].splitlines()
        )

    def test_log_close_to_bounds(self):
        """Test logging of variables and constraints near bounds."""
        m = self.build_model()

        with LoggingIntercept(None, 'pyomo.util.infeasible') as LOG:
            log_close_to_bounds(m)
        self.assertEqual(
            'log_close_to_bounds() called with a logger whose '
            'effective level is higher than logging.INFO: no output '
            'will be logged regardless of bound status',
            LOG.getvalue().strip(),
        )

        output = StringIO()
        with LoggingIntercept(output, 'pyomo.util.infeasible', logging.INFO):
            log_close_to_bounds(m)
        expected_output = [
            "y near UB of 2",
            "yy near LB of 0",
            "c4 near LB of 1.9999999",
            "c11 near UB of 1.9999999",
        ]
        self.assertEqual(expected_output, output.getvalue().splitlines())

    def test_log_infeasible_constraints_verbose_expressions(self):
        """Test for logging of infeasible constraints."""
        m = self.build_model()
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.util.infeasible', logging.INFO):
            log_infeasible_constraints(m, log_expression=True)
        expected_output = [
            "CONSTR c1: 2.0 </= 1",
            "  - EXPR: 2.0 </= x",
            "CONSTR c2: 1 =/= 4.0",
            "  - EXPR: x =/= 4.0",
            "CONSTR c3: 1 </= 0.0",
            "  - EXPR: x </= 0.0",
            "CONSTR c5: 5.0 <?= evaluation error <?= 10.0",
            "  - EXPR: 5.0 <?= z <?= 10.0",
            "CONSTR c7: evaluation error =?= 6.0",
            "  - EXPR: z =?= 6.0",
            "CONSTR c8: 3.0 </= 1 <= 6.0",
            "  - EXPR: 3.0 </= x <= 6.0",
            "CONSTR c9: 0.0 <= 1 </= 0.5",
            "  - EXPR: 0.0 <= x </= 0.5",
        ]
        self.assertEqual(expected_output, output.getvalue().splitlines())

    def test_log_infeasible_constraints_verbose_variables(self):
        """Test for logging of infeasible constraints."""
        m = self.build_model()
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.util.infeasible', logging.INFO):
            log_infeasible_constraints(m, log_variables=True)
        expected_output = [
            "CONSTR c1: 2.0 </= 1",
            "  - VAR x: 1",
            "CONSTR c2: 1 =/= 4.0",
            "  - VAR x: 1",
            "CONSTR c3: 1 </= 0.0",
            "  - VAR x: 1",
            "CONSTR c5: 5.0 <?= evaluation error <?= 10.0",
            "  - VAR z: None",
            "CONSTR c7: evaluation error =?= 6.0",
            "  - VAR z: None",
            "CONSTR c8: 3.0 </= 1 <= 6.0",
            "  - VAR x: 1",
            "CONSTR c9: 0.0 <= 1 </= 0.5",
            "  - VAR x: 1",
        ]
        self.assertEqual(expected_output, output.getvalue().splitlines())


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import json
import sys
import argparse

from pyomo.opt import SolverFactory, UnknownSolver
from pyomo.scripting.pyomo_parser import add_subparser, CustomHelpFormatter


def create_parser(parser=None):
    #
    # Setup command-line options.  The '--solver' option creates
    # all subsequent options...
    #
    if parser is None:
        parser = argparse.ArgumentParser(
            usage='%(prog)s [options] <model_or_config_file> [<data_files>]'
        )
    parser.add_argument('--solver', action='store', dest='solver', default=None)
    parser.add_argument(
        '--solver-manager', action='store', dest='solver_manager', default='serial'
    )
    parser.add_argument(
        '--generate-config-template', action='store', dest='template', default=None
    )
    return parser


def create_temporary_parser(solver=False, generate=False):
    #
    # We create a dummy parser and subparser, to help make the 'help'
    # output seem sensible.
    #
    parser = argparse.ArgumentParser(formatter_class=CustomHelpFormatter)
    _subparsers = parser.add_subparsers()
    _parser = _subparsers.add_parser('solve')
    _parser.formatter_class = CustomHelpFormatter
    if generate:
        #
        # Adding documentation about the two options that are
        # defined in the initial parser.
        #
        _parser.add_argument(
            '--generate-config-template',
            action='store',
            dest='template',
            default=None,
            help='Create a configuration template file in YAML or JSON and exit.',
        )
    if solver:
        _parser.add_argument(
            '--solver',
            action='store',
            dest='solver',
            default=None,
            help='Solver name. This option is required unless the solver name is'
            'specified in a configuration file.',
        )
        _parser.usage = '%(prog)s [options] <model_or_config_file> [<data_files>]'
        _parser.epilog = """
Description:

The 'pyomo solve' subcommand optimizes a Pyomo model.  The --solver option
is required because the specific steps executed are solver dependent.
The standard steps executed by this subcommand are:

  - Apply pre-solve operations (e.g. import Python packages)
  - Create the model
  - Apply model transformations
  - Perform optimization
  - Apply post-solve operations (e.g. save optimal solutions)


This subcommand can be executed with or without a configuration file.
The command-line options can be used to perform simple optimization steps.
For example:

  pyomo solve --solver=glpk model.py model.dat

This uses the 'glpk' solver to optimize the model define in 'model.py'
using the Pyomo data file 'pyomo.dat'.  Solver-specific command-line
options can be listed by specifying the '--solver' option and the '--help'
(or '-h') option:

  pyomo solve --solver=glpk --help


A yaml or json configuration file can also be used to specify
options used by the solver.  For example:

  pyomo solve --solver=glpk config.yaml

No other command-line options are required!  Further, the '--solver'
option can be omitted if the solver name is specified in the configuration
file.

Configuration options are also solver-specific; a template configuration
file can be generated with the '--generate-config-template' option:

  pyomo solve --solver=glpk --generate-config-template=template.yaml
  pyomo solve --solver=glpk --generate-config-template=template.json

Note that yaml template files contain comments that describe the
configuration options.  Also, configuration files will generally support
more configuration options than are available with command-line options.

"""
    #
    _parser.add_argument(
        'model_or_config_file',
        action='store',
        nargs='?',
        default='',
        help="A Python module that defines a Pyomo model, or a configuration file "
        "that defines options for 'pyomo solve' (in either YAML or JSON format)",
    )
    _parser.add_argument(
        'data_files',
        action='store',
        nargs='*',
        default=[],
        help='Pyomo data files that defined data used to initialize the model '
        '(specified in the first argument)',
    )
    #
    return _parser


def solve_exec(args, unparsed):
    import pyomo.scripting.util

    #
    solver_manager = getattr(args, 'solver_manager', None)
    solver = getattr(args, 'solver', None)
    #
    if solver is None:
        #
        # Get configuration values if no solver has been specified
        #
        try:
            val = pyomo.scripting.util.get_config_values(unparsed[-1])
        except IndexError:
            val = None
        except IOError:
            val = None
        #
        # Try to get the solver name
        #
        if not val is None:
            try:
                solver = val['solvers'][0]['solver name']
            except:
                solver = None
        #
        # If no luck, then generate an error
        #
        if solver is None:
            if not ('-h' in unparsed or '--help' in unparsed):
                print("ERROR: No solver specified!")
                print("")
            parser = create_temporary_parser(solver=True, generate=True)
            parser.parse_args(args=unparsed + ['-h'])
            sys.exit(1)

    config = None
    _solver = solver if solver_manager == 'serial' else 'asl'
    with SolverFactory(_solver) as opt:
        if opt is None or isinstance(opt, UnknownSolver):
            print("ERROR: Unknown solver '%s'!" % solver)
            sys.exit(1)
        #
        # Generate a template file
        #
        if not args.template is None:
            config = opt.config_block(init=True)
            OUTPUT = open(args.template, 'w')
            if args.template.endswith('json'):
                OUTPUT.write(json.dumps(config.value(), indent=2))
            else:
                OUTPUT.write(config.generate_yaml_template())
            OUTPUT.close()
            print("  Created template file '%s'" % args.template)
            sys.exit(0)
        #
        # Parse previously unparsed options
        #
        config = opt.config_block()
        if '-h' in unparsed or '--help' in unparsed:
            _parser = create_temporary_parser(generate=True)
        else:
            _parser = create_temporary_parser()
        config.initialize_argparse(_parser)
        _parser.usage = '%(prog)s [options] <model_or_config_file> [<data_files>]'
        _options = _parser.parse_args(args=unparsed)
        #
        # Import the parsed values into the config block, then
        # create an Options object
        #
        config.import_argparse(_options)
        config.solvers[0].solver_name = getattr(args, 'solver', None)
        if _options.model_or_config_file.endswith('.py'):
            config.model.filename = _options.model_or_config_file
            config.data.files = _options.data_files
        else:
            val = pyomo.scripting.util.get_config_values(_options.model_or_config_file)
            config.set_value(val)

    if config is None:
        raise RuntimeError("Failed to create config object")
    config.solvers[0].solver_name = solver
    config.solvers[0].manager = solver_manager

    from pyomo.scripting.pyomo_command import run_pyomo

    #
    # Note that we pass-in pre-parsed options.  The run_command()
    # function knows to not perform a parse, but instead to simply
    # used these parsed values.
    #
    return pyomo.scripting.util.run_command(
        command=run_pyomo, parser=_parser, options=config, name='pyomo solve'
    )


#
# Add a subparser for the solve command
#
solve_parser = create_parser(
    add_subparser(
        'solve',
        func=solve_exec,
        help='Optimize a model',
        add_help=False,
        description='This pyomo subcommand is used to analyze optimization models.',
    )
)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for BuildCheck() Objects
#
# PyomoModel                Base test class
# Scalar                    Test scalar parameter
# Array1                    Test arrays of parameters
#

import os
from io import StringIO

import pyomo.common.unittest as unittest

from pyomo.environ import AbstractModel, BuildCheck, Param, Set, value


class PyomoModel(unittest.TestCase):
    def setUp(self):
        self.model = AbstractModel()
        self.instance = None

    def tearDown(self):
        self.model = None
        self.instance = None

    def construct(self, filename):
        self.instance = self.model.create_instance(filename)


def action1a_fn(model):
    return value(model.A) == 3.3


def action1b_fn(model):
    return value(model.A) != 3.3


def action2a_fn(model, i):
    ans = True
    if i in model.A:
        return value(model.A[i]) == 1.3
    return True


def action2b_fn(model, i):
    if i in model.A:
        ans = value(model.A[i]) == 1.3
        # print "HERE",i,ans,not ans
        return not ans
    return True


class Scalar(PyomoModel):
    def setUp(self):
        #
        # Create Model
        #
        PyomoModel.setUp(self)
        #
        # Create model instance
        #
        self.model.A = Param(initialize=3.3)

    def tearDown(self):
        if os.path.exists("param.dat"):
            os.remove("param.dat")
        PyomoModel.tearDown(self)

    def test_true(self):
        """Apply a build check that returns true"""
        self.model.action1 = BuildCheck(rule=action1a_fn)
        self.instance = self.model.create_instance()
        tmp = value(self.instance.A)
        self.assertEqual(tmp, 3.3)

    def test_false(self):
        """Apply a build check that returns false"""
        self.model.action1 = BuildCheck(rule=action1b_fn)
        try:
            self.instance = self.model.create_instance()
            self.fail("expected failure")
        except ValueError:
            pass


class Array1(PyomoModel):
    def setUp(self):
        #
        # Create Model
        #
        PyomoModel.setUp(self)
        #
        # Create model instance
        #
        self.model.Z = Set(initialize=[1, 3])
        self.model.A = Param(self.model.Z, initialize=1.3)

    def tearDown(self):
        PyomoModel.tearDown(self)

    def test_true(self):
        """Check the value of the parameter"""
        self.model.action2 = BuildCheck(self.model.Z, rule=action2a_fn)
        self.instance = self.model.create_instance()

    def test_false(self):
        """Check the value of the parameter"""
        self.model.action2 = BuildCheck(self.model.Z, rule=action2b_fn)
        try:
            self.instance = self.model.create_instance()
            self.fail("expected failure")
        except ValueError:
            pass


class Array2(PyomoModel):
    def setUp(self):
        #
        # Create Model
        #
        PyomoModel.setUp(self)
        #
        # Create model instance
        #
        self.model.Z = Set(initialize=[1, 3])
        self.model.A = Param(self.model.Z, initialize=1.3)

    def tearDown(self):
        PyomoModel.tearDown(self)

    def test_true(self):
        """Check the value of the parameter"""
        self.model.action2 = BuildCheck(self.model.Z, rule=action2a_fn)
        self.instance = self.model.create_instance()

    def test_false(self):
        """Check the value of the parameter"""
        self.model.action2 = BuildCheck(self.model.Z, rule=action2b_fn)
        try:
            self.instance = self.model.create_instance()
            self.fail("expected failure")
        except ValueError:
            pass


class TestMisc(unittest.TestCase):
    def test_error1(self):
        model = AbstractModel()
        try:
            model.a = BuildCheck()
            self.fail("Expected ValueError")
        except ValueError:
            pass

    def test_io(self):
        model = AbstractModel()
        model.c1 = BuildCheck(rule=lambda M: True)
        model.A = Set(initialize=[1, 2, 3])
        model.c2 = BuildCheck(model.A, rule=lambda M, i: True)
        instance = model.create_instance()
        #
        buf = StringIO()
        instance.pprint(ostream=buf)
        self.assertEqual(
            buf.getvalue(),
            """1 Set Declarations
    A : Size=1, Index=None, Ordered=Insertion
        Key  : Dimen : Domain : Size : Members
        None :     1 :    Any :    3 : {1, 2, 3}

2 BuildCheck Declarations
    c1 : 
    c2 : 

3 Declarations: c1 A c2
""",
        )


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.expr.numeric_expr import (
    NegationExpression,
    ProductExpression,
    DivisionExpression,
    PowExpression,
    Expr_ifExpression,
    mutable_expression,
)
import pyomo.repn.linear as linear
import pyomo.repn.util as util
from pyomo.repn.linear import _merge_dict, to_expression
from pyomo.repn.util import sum_like_expression_types, val2str, InvalidNumber

_CONSTANT = linear.ExprType.CONSTANT
_FIXED = linear.ExprType.FIXED
_LINEAR = linear.ExprType.LINEAR
_GENERAL = linear.ExprType.GENERAL
_QUADRATIC = linear.ExprType.QUADRATIC


class QuadraticRepn:
    __slots__ = ("multiplier", "constant", "linear", "quadratic", "nonlinear")

    def __init__(self):
        self.multiplier = 1
        self.constant = 0
        self.linear = {}
        self.quadratic = None
        self.nonlinear = None

    def __str__(self):
        linear = (
            "{"
            + ", ".join(f"{val2str(k)}: {val2str(v)}" for k, v in self.linear.items())
            + "}"
        )
        if self.quadratic is None:
            quadratic = None
        else:
            quadratic = (
                "{"
                + ", ".join(
                    f"({val2str(k1)}, {val2str(k2)}): {val2str(v)}"
                    for (k1, k2), v in self.quadratic.items()
                )
                + "}"
            )
        return (
            f"{self.__class__.__name__}(mult={val2str(self.multiplier)}, "
            f"const={val2str(self.constant)}, "
            f"linear={linear}, quadratic={quadratic}, "
            f"nonlinear={val2str(self.nonlinear)})"
        )

    def __repr__(self):
        return str(self)

    @staticmethod
    def constant_flag(val):
        return val

    @staticmethod
    def multiplier_flag(val):
        return val

    def walker_exitNode(self):
        if self.nonlinear is not None:
            return _GENERAL, self
        elif self.quadratic:
            return _QUADRATIC, self
        elif self.linear:
            return _LINEAR, self
        else:
            return _CONSTANT, self.multiplier * self.constant

    def duplicate(self):
        ans = self.__class__.__new__(self.__class__)
        ans.multiplier = self.multiplier
        ans.constant = self.constant
        ans.linear = dict(self.linear)
        if self.quadratic:
            ans.quadratic = dict(self.quadratic)
        else:
            ans.quadratic = None
        ans.nonlinear = self.nonlinear
        return ans

    def to_expression(self, visitor):
        var_map = visitor.var_map
        if self.nonlinear is not None:
            # We want to start with the nonlinear term (and use
            # assignment) in case the term is a non-numeric node (like a
            # relational expression)
            ans = self.nonlinear
        else:
            ans = 0
        with mutable_expression() as e:
            if self.quadratic:
                for (x1, x2), coef in self.quadratic.items():
                    if self.multiplier_flag(coef):
                        if x1 == x2:
                            e += coef * var_map[x1] ** 2
                        else:
                            e += coef * (var_map[x1] * var_map[x2])
            if self.linear:
                var_map = visitor.var_map
                for vid, coef in self.linear.items():
                    if self.multiplier_flag(coef):
                        e += coef * var_map[vid]
            if self.constant_flag(self.constant):
                e += self.constant
        if e.nargs() > 1:
            ans += e
        elif e.nargs() == 1:
            ans += e.arg(0)
        if self.multiplier_flag(self.multiplier) != 1:
            ans *= self.multiplier
        return ans

    def append(self, other):
        """Append a child result from acceptChildResult

        Notes
        -----
        This method assumes that the operator was "+". It is implemented
        so that we can directly use a QuadraticRepn() as a data object in
        the expression walker (thereby avoiding the function call for a
        custom callback)

        """
        # Note that self.multiplier will always be 1 (we only call append()
        # within a sum, so there is no opportunity for self.multiplier to
        # change). Omitting the assertion for efficiency.
        # assert self.multiplier == 1
        _type, other = other
        if _type <= _FIXED:
            self.constant += other
            return

        other_mult_flag = self.multiplier_flag(other.multiplier)
        if other_mult_flag == 1:
            self.constant += other.constant
            if other.linear:
                l = self.linear
                for vid, coef in other.linear.items():
                    if vid in l:
                        l[vid] += coef
                    else:
                        l[vid] = coef
            if other.quadratic:
                if not self.quadratic:
                    self.quadratic = other.quadratic
                else:
                    q = self.quadratic
                    for vid, coef in other.quadratic.items():
                        if vid in q:
                            q[vid] += coef
                        else:
                            q[vid] = coef
            if other.nonlinear is not None:
                if self.nonlinear is None:
                    self.nonlinear = other.nonlinear
                else:
                    self.nonlinear += other.nonlinear
            return

        mult = other.multiplier
        if not other_mult_flag:
            # 0 * other, so you would think that there is nothing to
            # add/change about self.  However, there is a chance
            # that other contains an InvalidNumber, so we should go
            # looking for it...
            if other.constant.__class__ is InvalidNumber:
                self.constant += mult * other.constant
            for vid, coef in other.linear.items():
                if coef.__class__ is InvalidNumber:
                    if vid in self.linear:
                        self.linear[vid] += mult * coef
                    else:
                        self.linear[vid] = mult * coef
            if other.quadratic:
                for vid, coef in other.quadratic.items():
                    if coef.__class__ is InvalidNumber:
                        if not self.quadratic:
                            self.quadratic = {}
                        if vid in self.quadratic:
                            self.quadratic[vid] += mult * coef
                        else:
                            self.quadratic[vid] = mult * coef
        else:
            #
            # other.multiplier other than 0 or 1:
            #
            if self.constant_flag(other.constant):
                self.constant += mult * other.constant
            if other.linear:
                l = self.linear
                for vid, coef in other.linear.items():
                    if vid in l:
                        l[vid] += mult * coef
                    else:
                        l[vid] = mult * coef
            if other.quadratic:
                if self.quadratic:
                    q = self.quadratic
                else:
                    q = self.quadratic = {}
                for vid, coef in other.quadratic.items():
                    if vid in q:
                        q[vid] += mult * coef
                    else:
                        q[vid] = mult * coef

        if other.nonlinear is not None:
            if self.nonlinear is None:
                self.nonlinear = mult * other.nonlinear
            else:
                self.nonlinear += mult * other.nonlinear


def _mul_linear_linear(visitor, linear1, linear2):
    vo = visitor.var_recorder.var_order
    quadratic = {}
    l2 = [(k, v, vo[k]) for k, v in linear2.items()]
    for vid1, coef1 in linear1.items():
        o1 = vo[vid1]
        for vid2, coef2, o2 in l2:
            if o1 < o2:
                key = vid1, vid2
            else:
                key = vid2, vid1
            if key in quadratic:
                quadratic[key] += coef1 * coef2
            else:
                quadratic[key] = coef1 * coef2
    return quadratic


def _handle_product_linear_linear(visitor, node, arg1, arg2):
    # We are multiplying (A + Bx) * (A + Bx)
    _, arg1 = arg1
    _, arg2 = arg2
    # [BB]: Quadratic first, because we will update linear in a minute
    arg1.quadratic = _mul_linear_linear(visitor, arg1.linear, arg2.linear)
    # Linear second, as this relies on knowing the original constants
    # [BA]
    arg1_const_flag = arg1.constant_flag(arg1.constant)
    arg2_const_flag = arg2.constant_flag(arg2.constant)
    if arg2_const_flag == 1:
        pass
    elif not arg2_const_flag:
        # linear * 0, so you would think there is nothing to do, but we
        # need to preserve any InvalidNumbers
        arg1.linear = {
            vid: coef * arg2.constant
            for key, coef in arg1.linear.items()
            if coef.__class__ is InvalidNumber
        }
    else:
        c = arg2.constant
        l = arg1.linear
        for vid, coef in l.items():
            l[vid] = c * coef
    # [AB]
    _merge_dict(arg1.linear, arg2.linear, arg1.constant, arg1_const_flag)
    # [AA]: Finally, the constant and multipliers
    if arg1_const_flag and arg2_const_flag:
        arg1.constant *= arg2.constant
    else:
        arg1.constant = 0
    arg1.multiplier *= arg2.multiplier
    return _QUADRATIC, arg1


def _handle_product_nonlinear(visitor, node, arg1, arg2):
    ans = visitor.Result()
    # We are multiplying (A + Bx + Cx^2 + D(x)) * (A + Bx + Cx^2 + Dx))
    _t1, x1 = arg1
    _t2, x2 = arg2
    if not visitor.expand_nonlinear_products:
        ans.nonlinear = to_expression(visitor, arg1) * to_expression(visitor, arg2)
        return _GENERAL, ans

    ans.multiplier = x1.multiplier * x2.multiplier
    x1.multiplier = x2.multiplier = 1
    # x1.const * x2.const [AA]
    ans.constant = x1.constant * x2.constant
    # linear & quadratic terms
    ans.quadratic = {}
    # [BA], [CA]
    x1_const_flag = ans.constant_flag(x1.constant)
    x2_const_flag = ans.constant_flag(x2.constant)
    _merge_dict(ans.linear, x1.linear, x2.constant, x2_const_flag)
    _merge_dict(ans.quadratic, x1.quadratic, x2.constant, x2_const_flag)
    # [AB], [AC]
    _merge_dict(ans.linear, x2.linear, x1.constant, x1_const_flag)
    _merge_dict(ans.quadratic, x2.quadratic, x1.constant, x1_const_flag)
    # [BB]
    if x1.linear and x2.linear:
        quad = _mul_linear_linear(visitor, x1.linear, x2.linear)
        if ans.quadratic:
            _merge_dict(ans.quadratic, quad, 1, 1)
        else:
            ans.quadratic = quad
    # [DA] + [DB] + [DC] + [DD]
    NL = 0
    if x1.nonlinear is not None:
        NL += x1.nonlinear * x2.to_expression(visitor)
    x1.nonlinear = None
    x2.constant = 0
    x1_c = x1.constant
    x1.constant = 0
    x1_lin = x1.linear
    x1.linear = {}
    # [CB] + [CC] + [CD]
    if x1.quadratic:
        NL += x1.to_expression(visitor) * x2.to_expression(visitor)
        x1.quadratic = None
    x2.linear = {}
    # [BC] + [BD]
    if x1_lin and (x2.nonlinear is not None or x2.quadratic):
        x1.linear = x1_lin
        NL += x1.to_expression(visitor) * x2.to_expression(visitor)
    # [AD]
    if x1_const_flag and x2.nonlinear is not None:
        NL += x1_c * x2.nonlinear
    if NL.__class__ in sum_like_expression_types and NL.nargs() == 1:
        NL = NL.arg(0)
    ans.nonlinear = NL
    return _GENERAL, ans


def _handle_pow_linear_constant(visitor, node, arg1, arg2):
    _, exp = arg2
    if exp == 2:
        _, ans = arg1
        ans.multiplier = ans.multiplier**2
        ans.quadratic = {(vid, vid): coef * coef for vid, coef in ans.linear.items()}
        if len(ans.linear) > 1:
            vo = visitor.var_recorder.var_order
            l = [(vid, coef, vo[vid]) for vid, coef in ans.linear.items()]
            for i, (x1, c1, o1) in enumerate(l):
                for j, (x2, c2, o2) in enumerate(l):
                    if i == j:
                        break
                    if o1 < o2:
                        ans.quadratic[x1, x2] = 2 * c1 * c2
                    else:
                        ans.quadratic[x2, x1] = 2 * c1 * c2
        if ans.constant_flag(ans.constant):
            c = 2 * ans.constant
            for vid in ans.linear:
                ans.linear[vid] *= c
            ans.constant = ans.constant**2
        else:
            ans.linear = {}
        return _QUADRATIC, ans
    return linear._handle_pow_ANY_constant(visitor, node, arg1, arg2)


def define_exit_node_handlers(_exit_node_handlers=None):
    if _exit_node_handlers is None:
        _exit_node_handlers = {}
    linear.define_exit_node_handlers(_exit_node_handlers)
    #
    # NEGATION
    #
    _exit_node_handlers[NegationExpression][(_QUADRATIC,)] = linear._handle_negation_ANY
    #
    # PRODUCT
    #
    _exit_node_handlers[ProductExpression].update(
        {
            None: _handle_product_nonlinear,
            (_CONSTANT, _QUADRATIC): linear._handle_product_constant_ANY,
            (_FIXED, _QUADRATIC): linear._handle_product_constant_ANY,
            (_QUADRATIC, _CONSTANT): linear._handle_product_ANY_constant,
            (_QUADRATIC, _FIXED): linear._handle_product_ANY_constant,
            # Replace handler from the linear walker
            (_LINEAR, _LINEAR): _handle_product_linear_linear,
        }
    )
    #
    # DIVISION
    #
    _exit_node_handlers[DivisionExpression].update(
        {
            (_QUADRATIC, _CONSTANT): linear._handle_division_ANY_constant,
            (_QUADRATIC, _FIXED): linear._handle_division_ANY_fixed,
        }
    )
    #
    # EXPONENTIATION
    #
    _exit_node_handlers[PowExpression].update(
        {
            (_LINEAR, _CONSTANT): _handle_pow_linear_constant,
            (_QUADRATIC, _CONSTANT): linear._handle_pow_ANY_constant,
        }
    )
    #
    # ABS and UNARY handlers
    #
    # (no changes needed)
    #
    # NAMED EXPRESSION handlers
    #
    # (no changes needed)
    #
    # EXPR_IF handlers
    #
    _exit_node_handlers[Expr_ifExpression].update(
        {
            (_CONSTANT, i, _QUADRATIC): linear._handle_expr_if_const
            for i in (_CONSTANT, _FIXED, _LINEAR, _QUADRATIC, _GENERAL)
        }
    )
    _exit_node_handlers[Expr_ifExpression].update(
        {
            (_CONSTANT, _QUADRATIC, i): linear._handle_expr_if_const
            for i in (_CONSTANT, _FIXED, _LINEAR, _GENERAL)
        }
    )
    #
    # RELATIONAL handlers
    #
    # (no changes needed)
    return _exit_node_handlers


class QuadraticRepnVisitor(linear.LinearRepnVisitor):
    Result = QuadraticRepn
    exit_node_dispatcher = linear.ExitNodeDispatcher(
        util.initialize_exit_node_dispatcher(define_exit_node_handlers())
    )
    max_exponential_expansion = 2

    def _filter_zeros(self, ans):
        _flag = ans.constant_flag
        # Note: creating the intermediate list is important, as we are
        # modifying the dict in place.
        for vid in [vid for vid, c in ans.linear.items() if not _flag(c)]:
            del ans.linear[vid]
        if ans.quadratic:
            for vid in [vid for vid, c in ans.quadratic.items() if not _flag(c)]:
                del ans.quadratic[vid]
            if not ans.quadratic:
                ans.quadratic = None

    def _factor_multiplier_into_ans(self, ans, mult):
        _flag = ans.constant_flag
        linear = ans.linear
        zeros = []
        for vid, coef in linear.items():
            prod = coef * mult
            if _flag(prod):
                linear[vid] = prod
            else:
                zeros.append(vid)
        for vid in zeros:
            del linear[vid]
        if ans.quadratic:
            quadratic = ans.quadratic
            zeros = []
            for vid, coef in quadratic.items():
                prod = coef * mult
                if _flag(prod):
                    quadratic[vid] = prod
                else:
                    zeros.append(vid)
            for vid in zeros:
                del quadratic[vid]
            if not quadratic:
                ans.quadratic = None
        if ans.nonlinear is not None:
            ans.nonlinear *= mult
        if _flag(ans.constant):
            ans.constant *= mult
        ans.multiplier = 1



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.core.expr as EXPR

from pyomo.core import (
    nonpyomo_leaf_types,
    TransformationFactory,
    IntegerSet,
    Integers,
    PositiveIntegers,
    NonPositiveIntegers,
    NegativeIntegers,
    NonNegativeIntegers,
    Reals,
    PositiveReals,
    NonNegativeReals,
    NegativeReals,
    NonPositiveReals,
    PercentFraction,
    RealSet,
    Var,
    Set,
    value,
    Binary,
    Constraint,
    Objective,
)
from pyomo.core.base.misc import create_name
from pyomo.core.plugins.transform.util import partial
from pyomo.core.plugins.transform.hierarchy import IsomorphicTransformation
from pyomo.core.plugins.transform.util import collectAbstractComponents


import logging

logger = logging.getLogger('pyomo.core')


class VarmapVisitor(EXPR.ExpressionReplacementVisitor):
    def __init__(self, varmap):
        super(VarmapVisitor, self).__init__()
        self.varmap = varmap

    def visiting_potential_leaf(self, node):
        if node.__class__ in nonpyomo_leaf_types:
            return True, node
        #
        # Clone leaf nodes in the expression tree
        #
        if node.is_variable_type():
            if node.local_name in self.varmap:
                return True, self.varmap[node.local_name]
            else:
                return True, node

        if isinstance(node, EXPR.LinearExpression):
            with EXPR.nonlinear_expression() as expr:
                for c, v in zip(node.linear_coefs, node.linear_vars):
                    if hasattr(v, 'local_name'):
                        expr += c * self.varmap.get(v.local_name)
                    else:
                        expr += c * v
            return True, expr

        return False, None


def _walk_expr(expr, varMap):
    """
    Walks an expression tree, making the replacements defined in varMap
    """
    visitor = VarmapVisitor(varMap)
    return visitor.dfs_postorder_stack(expr)


@TransformationFactory.register(
    "core.nonnegative_vars",
    doc="Create an equivalent model in which all variables lie in the nonnegative orthant.",
)
class NonNegativeTransformation(IsomorphicTransformation):
    """
    Creates a new, equivalent model by forcing all variables to lie in
    the nonnegative orthant by introducing auxiliary variables.
    """

    def __init__(self, **kwds):
        kwds["name"] = kwds.pop("name", "vars")
        super(NonNegativeTransformation, self).__init__(**kwds)

        self.realSets = (
            Reals,
            PositiveReals,
            NonNegativeReals,
            NegativeReals,
            NonPositiveReals,
            PercentFraction,
            RealSet,
        )
        # Intentionally leave out Binary, Boolean, BinarySet, and BooleanSet;
        # we check for those explicitly
        self.discreteSets = (
            IntegerSet,
            Integers,
            PositiveIntegers,
            NonPositiveIntegers,
            NegativeIntegers,
            NonNegativeIntegers,
        )

    def _create_using(self, model, **kwds):
        """
        Force all variables to lie in the nonnegative orthant.

        Required arguments:
            model       The model to transform.

        Optional keyword arguments:
          pos_suffix    The suffix applied to the 'positive' component of
                            converted variables. Default is '_plus'.
          neg_suffix    The suffix applied to the 'positive' component of
                            converted variables. Default is '_minus'.
        """
        #
        # Optional naming schemes
        #
        pos_suffix = kwds.pop("pos_suffix", "_plus")
        neg_suffix = kwds.pop("neg_suffix", "_minus")
        #
        # We first perform an abstract problem transformation. Then, if model
        # data is available, we instantiate the new model. If not, we construct
        # a mapping that can later be used to populate the new model.
        #
        nonneg = model.clone()
        components = collectAbstractComponents(nonneg)

        # Map from variable base names to a {index, rule} map
        constraint_rules = {}

        # Map from variable base names to a rule defining the domains for that
        # variable
        domain_rules = {}

        # Map from variable base names to its set of indices
        var_indices = {}

        # Map from fully qualified variable names to replacement expressions.
        # For now, it is actually a map from a variable name to a closure that
        # must later be evaluated with a model containing the replacement
        # variables.
        var_map = {}

        #
        # Get the constraints that enforce the bounds and domains of each
        # variable
        #
        for var_name in components["Var"]:
            var = nonneg.__getattribute__(var_name)

            # Individual bounds and domains
            orig_bounds = {}
            orig_domain = {}

            # New indices
            indices = set()

            # Map from constraint names to a constraint rule.
            constraints = {}

            # Map from variable indices to a domain
            domains = {}

            for ndx in var:
                # Fully qualified variable name
                vname = create_name(str(var_name), ndx)

                # We convert each index to a string to avoid difficult issues
                # regarding appending a suffix to tuples.
                #
                # If the index is None, this casts the index to a string,
                # which doesn't match up with how Pyomo treats None indices
                # internally. Replace with "" to be consistent.
                if ndx is None:
                    v_ndx = ""
                else:
                    v_ndx = str(ndx)

                # Get the variable bounds
                lb = value(var[ndx].lb)
                ub = value(var[ndx].ub)
                orig_bounds[ndx] = (lb, ub)

                # Get the variable domain
                if var[ndx].domain is not None:
                    orig_domain[ndx] = var[ndx].domain
                else:
                    orig_domain[ndx] = var.domain

                # Determine the replacement expression. Either a new single
                # variable with the same attributes, or a sum of two new
                # variables.
                #
                # If both the bounds and domain allow for negative values,
                # replace the variable with the sum of nonnegative ones.

                bounds_neg = (
                    orig_bounds[ndx] == (None, None)
                    or orig_bounds[ndx][0] is None
                    or orig_bounds[ndx][0] < 0
                )
                domain_neg = (
                    orig_domain[ndx] is None
                    or orig_domain[ndx].bounds()[0] is None
                    or orig_domain[ndx].bounds()[0] < 0
                )
                if bounds_neg and domain_neg:
                    # Make two new variables.
                    posVarSuffix = "%s%s" % (v_ndx, pos_suffix)
                    negVarSuffix = "%s%s" % (v_ndx, neg_suffix)

                    new_indices = (posVarSuffix, negVarSuffix)

                    # Replace the original variable with a sum expression
                    expr_dict = {posVarSuffix: 1, negVarSuffix: -1}
                else:
                    # Add the new index. Lie if is 'None', since Pyomo treats
                    # 'None' specially as a key.
                    #
                    # More lies: don't let a blank index exist. Replace it with
                    # '_'. I don't actually have a justification for this other
                    # than that allowing "" as a key will eventually almost
                    # certainly lead to a strange bug.
                    if v_ndx is None:
                        t_ndx = "None"
                    elif v_ndx == "":
                        t_ndx = "_"
                    else:
                        t_ndx = v_ndx

                    new_indices = (t_ndx,)

                    # Replace the original variable with a sum expression
                    expr_dict = {t_ndx: 1}

                # Add the new indices
                for x in new_indices:
                    indices.add(x)

                # Replace the original variable with an expression
                var_map[vname] = partial(self.sumRule, var_name, expr_dict)

                # Enforce bounds as constraints
                if orig_bounds[ndx] != (None, None):
                    cname = "%s_%s" % (vname, "bounds")
                    tmp = orig_bounds[ndx]
                    constraints[cname] = partial(
                        self.boundsConstraintRule, tmp[0], tmp[1], var_name, expr_dict
                    )

                # Enforce the bounds of the domain as constraints
                if orig_domain[ndx] != None:
                    cname = "%s_%s" % (vname, "domain_bounds")
                    tmp = orig_domain[ndx].bounds()
                    constraints[cname] = partial(
                        self.boundsConstraintRule, tmp[0], tmp[1], var_name, expr_dict
                    )

                # Domain will either be NonNegativeReals, NonNegativeIntegers,
                # or Binary. We consider Binary because some solvers may
                # optimize over binary variables.
                if var[ndx].is_continuous():
                    for x in new_indices:
                        domains[x] = NonNegativeReals
                elif var[ndx].is_binary():
                    for x in new_indices:
                        domains[x] = Binary
                elif var[ndx].is_integer():
                    for x in new_indices:
                        domains[x] = NonNegativeIntegers
                else:
                    logger.warning(
                        "Warning: domain '%s' not recognized, "
                        "defaulting to 'NonNegativeReals'" % (var.domain,)
                    )
                    for x in new_indices:
                        domains[x] = NonNegativeReals

            constraint_rules[var_name] = constraints
            domain_rules[var_name] = partial(self.exprMapRule, domains)
            var_indices[var_name] = indices

        # Remove all existing variables.
        toRemove = []
        for attr_name, attr in nonneg.__dict__.items():
            if isinstance(attr, Var):
                toRemove.append(attr_name)
        for attr_name in toRemove:
            nonneg.__delattr__(attr_name)

        # Add the sets defining the variables, then the variables
        for k, v in var_indices.items():
            sname = "%s_indices" % k
            nonneg.__setattr__(sname, Set(initialize=v))
            nonneg.__setattr__(
                k,
                Var(
                    nonneg.__getattribute__(sname),
                    domain=domain_rules[k],
                    bounds=(0, None),
                ),
            )

        # Construct the model to get the variables and their indices
        # recognized in the model
        ##nonneg = nonneg.create()

        # Safe to evaluate the modifiedVars mapping
        for var in var_map:
            var_map[var] = var_map[var](nonneg)

        # Map from constraint base names to maps from indices to expressions
        constraintExprs = {}

        #
        # Convert all modified variables in all constraints in the original
        # problem
        #
        for conName in components["Constraint"]:
            con = nonneg.__getattribute__(conName)

            # Map from constraint indices to a corrected expression
            exprMap = {}

            for ndx, cdata in con._data.items():
                lower = _walk_expr(cdata.lower, var_map)
                body = _walk_expr(cdata.body, var_map)
                upper = _walk_expr(cdata.upper, var_map)

                # Lie if ndx is None. Pyomo treats 'None' indices specially.
                if ndx is None:
                    ndx = "None"

                # Cast indices to strings, otherwise tuples ruin everything
                exprMap[str(ndx)] = (lower, body, upper)

            # Add to list of expression maps
            constraintExprs[conName] = exprMap

        # Map from constraint base names to maps from indices to expressions
        objectiveExprs = {}

        #
        # Convert all modified variables in all objectives in the original
        # problem
        #
        for objName in components["Objective"]:
            obj = nonneg.__getattribute__(objName)

            # Map from objective indices to a corrected expression
            exprMap = {}

            for ndx, odata in obj._data.items():
                exprMap[ndx] = _walk_expr(odata.expr, var_map)

            # Add to list of expression maps
            objectiveExprs[objName] = exprMap

        # Make the modified original constraints
        for conName, ruleMap in constraintExprs.items():
            # Make the set of indices
            sname = conName + "_indices"
            _set = Set(initialize=ruleMap.keys())
            nonneg.__setattr__(sname, _set)
            _set.construct()

            # Define the constraint
            _con = Constraint(
                nonneg.__getattribute__(sname), rule=partial(self.exprMapRule, ruleMap)
            )
            nonneg.__setattr__(conName, _con)
            _con.construct()

        # Make the bounds constraints
        for varName, ruleMap in constraint_rules.items():
            conName = varName + "_constraints"
            # Make the set of indices
            sname = conName + "_indices"
            _set = Set(initialize=ruleMap.keys())
            nonneg.__setattr__(sname, _set)
            _set.construct()

            # Define the constraint
            _con = Constraint(
                nonneg.__getattribute__(sname),
                rule=partial(self.delayedExprMapRule, ruleMap),
            )
            nonneg.__setattr__(conName, _con)
            _con.construct()

        # Make the objectives
        for objName, ruleMap in objectiveExprs.items():
            # Make the set of indices
            sname = objName + "_indices"
            _set = Set(initialize=ruleMap.keys())
            nonneg.__setattr__(sname, _set)
            _set.construct()

            # Define the constraint
            _obj = Objective(
                nonneg.__getattribute__(sname), rule=partial(self.exprMapRule, ruleMap)
            )
            nonneg.__setattr__(objName, _obj)
            _obj.construct()

        return nonneg

    @staticmethod
    def boundsConstraintRule(lb, ub, attr, vars, model):
        """
        Produces 'lb < x^+ - x^- < ub' style constraints. Designed to
        be made a closer through functools.partial, across lb, ub, attr,
        and vars. vars is a {varname: coefficient} dictionary. attr is the
        base variable name; that is, X[1] would be referenced by

          model.__getattribute__('X')[1]

        and so attr='X', and 1 is a key of vars.

        """
        return (
            lb,
            sum(c * model.__getattribute__(attr)[v] for (v, c) in vars.items()),
            ub,
        )

    @staticmethod
    def noConstraint(*args):
        return None

    @staticmethod
    def sumRule(attr, vars, model):
        """
        Returns a sum expression.
        """
        return sum(c * model.__getattribute__(attr)[v] for (v, c) in vars.items())

    @staticmethod
    def exprMapRule(ruleMap, model, ndx=None):
        """Rule intended to return expressions from a lookup table"""
        return ruleMap[ndx]

    @staticmethod
    def delayedExprMapRule(ruleMap, model, ndx=None):
        """
        Rule intended to return expressions from a lookup table. Each entry
        in the lookup table is a functor that needs to be evaluated before
        returning.
        """
        return ruleMap[ndx](model)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
from pyomo.common.config import ConfigDict, ConfigValue


def get_configuration(config):
    ans = ConfigDict()
    ans.declare('key1', ConfigValue(default=0, domain=int))
    ans.declare('key2', ConfigValue(default=5, domain=str))
    return ans(config)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#

from pyomo.common.dependencies import attempt_import, UnavailableClass

cplex, cplex_available = attempt_import('cplex')


class IncumbentCallback_cplex(
    cplex.callbacks.IncumbentCallback if cplex_available else UnavailableClass(cplex)
):
    """Inherent class in Cplex to call Incumbent callback."""

    def __call__(self):
        """
        This is an inherent function in LazyConstraintCallback in CPLEX.
        This callback will be used after each new potential incumbent is found.
        https://www.ibm.com/support/knowledgecenter/SSSA5P_12.10.0/ilog.odms.cplex.help/refpythoncplex/html/cplex.callbacks.IncumbentCallback-class.html
        IncumbentCallback will be activated after Lazyconstraint callback, when the potential incumbent solution is satisfies the lazyconstraints.
        TODO: need to handle GOA same integer combination check in lazyconstraint callback in single_tree.py
        """
        mindtpy_solver = self.mindtpy_solver
        opt = self.opt
        config = self.config
        if config.single_tree:
            self.reject()
        else:
            temp = []
            for var in mindtpy_solver.mip.MindtPy_utils.discrete_variable_list:
                value = self.get_values(opt._pyomo_var_to_solver_var_map[var])
                temp.append(int(round(value)))
            mindtpy_solver.curr_int_sol = tuple(temp)

            if mindtpy_solver.curr_int_sol in set(mindtpy_solver.integer_list):
                self.reject()
